# pylint: disable=unnecessary-lambda,missing-docstring
"""coBib's configuration.

This file contains both, the actual implementation of the `Config` classes, as well as the runtime
`config` object, which gets exposed on the module level as `cobib.config.config`.
Note, that this last link will not point to the correct location in the online documentation due to
the nature of the lower-level import.
"""

from __future__ import annotations

import importlib.util
import io
import logging
import os
import sys
from abc import abstractmethod
from dataclasses import MISSING, dataclass, field
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, TextIO, Union

from typing_extensions import override

from cobib.utils.rel_path import RelPath

if TYPE_CHECKING:
    from .event import Event

LOGGER = logging.getLogger(__name__)


@dataclass
class _ConfigBase:
    """Base class for configuration section dataclasses."""

    @staticmethod
    def _assert(expression: bool, error: str) -> None:
        """Asserts the expression is True.

        Args:
            expression: the expression to assert.
            error: the message of the `RuntimeError` upon assertion failure.

        Raises:
            RuntimeError with the specified error string.
        """
        if not expression:
            raise RuntimeError(error)

    @abstractmethod
    def validate(self) -> None:
        """Validates the configuration at runtime.

        Raises:
            RuntimeError when an invalid setting is encountered.
        """

    def defaults(self) -> None:
        """Resets the configuration to the default settings."""
        # pylint: disable=no-member
        for name, field_ in self.__dataclass_fields__.items():
            if field_.default != MISSING:
                setattr(self, name, field_.default)
            else:
                setattr(self, name, field_.default_factory())  # type: ignore[misc]


@dataclass
class LoggingConfig(_ConfigBase):
    """The `config.logging` section."""

    cache: str | Path = "~/.cache/cobib/cache"
    """Specifies the default cache location."""
    logfile: str | Path = "~/.cache/cobib/cobib.log"
    """Specifies the default logfile location."""
    version: str | None = "~/.cache/cobib/version"
    """Specifies the location for the cached version number based on which coBib shows the latest
    changes. Set this to `None` to disable this functionality entirely."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the LOGGING configuration section.")
        self._assert(
            isinstance(self.cache, (str, Path)), "config.logging.cache should be a string."
        )
        self._assert(
            isinstance(self.logfile, (str, Path)), "config.logging.logfile should be a string."
        )
        self._assert(
            self.version is None or isinstance(self.version, str),
            "config.logging.version should be a string or `None`.",
        )


@dataclass
class AddCommandConfig(_ConfigBase):
    """The `config.commands.add` section."""

    skip_download: bool = False
    """Specifies whether to skip the attempt of downloading PDF files of added entries."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.ADD configuration section.")
        self._assert(
            isinstance(self.skip_download, bool),
            "config.commands.add.skip_download should be a boolean.",
        )


@dataclass
class DeleteCommandConfig(_ConfigBase):
    """The `config.commands.delete` section."""

    confirm: bool = True
    """Whether or not to confirm before deleting an entry."""

    preserve_files: bool = False
    """Specifies whether associated files should be preserved when deleting an entry."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.DELETE configuration section.")
        self._assert(
            isinstance(self.confirm, bool),
            "config.commands.delete.confirm should be a boolean.",
        )
        self._assert(
            isinstance(self.preserve_files, bool),
            "config.commands.delete.preserve_files should be a boolean.",
        )


@dataclass
class EditCommandConfig(_ConfigBase):
    """The `config.commands.edit` section."""

    default_entry_type: str = "article"
    """Specifies the default bibtex entry type."""
    editor: str = os.environ.get("EDITOR", "vim")
    """Specifies the editor program. Note, that this default will respect your `$EDITOR`
    environment setting and fall back to `vim` if that variable is not set."""
    preserve_files: bool = False
    """Specifies whether associated files should be preserved when renaming an entry during
    editing."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.EDIT configuration section.")
        self._assert(
            isinstance(self.default_entry_type, str),
            "config.commands.edit.default_entry_type should be a string.",
        )
        self._assert(
            isinstance(self.editor, str),
            "config.commands.edit.editor should be a string.",
        )
        self._assert(
            isinstance(self.preserve_files, bool),
            "config.commands.edit.preserve_files should be a boolean.",
        )


@dataclass
class ImportCommandConfig(_ConfigBase):
    """The `config.commands.import` section."""

    skip_download: bool = False
    """Specifies whether to skip the attempt of downloading PDF files of imported entries."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.IMPORT configuration section.")
        self._assert(
            isinstance(self.skip_download, bool),
            "config.commands.import.skip_download should be a boolean.",
        )


@dataclass
class ListCommandConfig(_ConfigBase):
    """The `config.commands.list_` section."""

    default_columns: list[str] = field(default_factory=lambda: ["label", "title"])
    """Specifies the default columns shown during the `list` command."""
    ignore_case: bool = False
    """Specifies whether filter matching should be performed case-insensitive."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.LIST configuration section.")
        self._assert(
            isinstance(self.default_columns, list),
            "config.commands.list_.default_columns should be a list.",
        )
        self._assert(
            isinstance(self.ignore_case, bool),
            "config.commands.list_.ignore_case should be a boolean.",
        )


@dataclass
class ModifyCommandConfig(_ConfigBase):
    """The `config.commands.modify` section."""

    preserve_files: bool = False
    """Specifies whether associated files should be preserved when renaming an entry during
    modifying."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.MODIFY configuration section.")
        self._assert(
            isinstance(self.preserve_files, bool),
            "config.commands.modify.preserve_files should be a boolean.",
        )


@dataclass
class OpenCommandConfig(_ConfigBase):
    """The `config.commands.open` section."""

    command: str = "xdg-open" if sys.platform.lower() == "linux" else "open"
    """Specifies the command used for opening files associated with your entries."""
    fields: list[str] = field(default_factory=lambda: ["file", "url"])
    """Specifies the entry fields which are to be checked for openable URLs."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.OPEN configuration section.")
        self._assert(
            isinstance(self.command, str),
            "config.commands.open.command should be a string.",
        )
        self._assert(
            isinstance(self.fields, list),
            "config.commands.open.fields should be a list.",
        )


@dataclass
class SearchHighlightConfig(_ConfigBase):
    """The `config.commands.search.highlights` section."""

    label: str = "blue"
    """Specifies the color with which to highlight the labels of search results."""
    query: str = "red"
    """Specifies the color with which to highlight the query matches of a search."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.SEARCH.HIGHLIGHTS configuration section.")
        self._assert(
            isinstance(self.label, str),
            "config.commands.search.highlights.label should be a string.",
        )
        self._assert(
            isinstance(self.query, str),
            "config.commands.search.highlights.query should be a string.",
        )


@dataclass
class SearchCommandConfig(_ConfigBase):
    """The `config.commands.search` section."""

    context: int = 1
    """Specifies the default number of context line to provide for each search query match. This is
    similar to the `-C` option of `grep`."""
    grep: str = "grep"
    """Specifies the grep tool used for searching through your database and associated files. The
    default tool (`grep`) will not provide results for attached PDFs but other tools such as
    [ripgrep-all](https://github.com/phiresky/ripgrep-all) will."""
    grep_args: list[str] = field(default_factory=list)
    """Specifies additional arguments for your grep command. Note, that GNU's grep understands
    extended regex patterns even without specifying `-E`."""
    ignore_case: bool = False
    """Specifies whether searches should be performed case-insensitive."""
    highlights: SearchHighlightConfig = field(default_factory=lambda: SearchHighlightConfig())
    """The nested section for highlights used when displaying search results."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.SEARCH configuration section.")
        self._assert(
            isinstance(self.context, int) and self.context >= 0,
            "config.commands.search.context should be a non-negative integer.",
        )
        self._assert(
            isinstance(self.grep, str),
            "config.commands.search.grep should be a string.",
        )
        self._assert(
            isinstance(self.grep_args, list),
            "config.commands.search.grep_args should be a list.",
        )
        self._assert(
            isinstance(self.ignore_case, bool),
            "config.commands.search.ignore_case should be a boolean.",
        )
        self.highlights.validate()


@dataclass
class CommandConfig(_ConfigBase):
    """The `config.commands` section."""

    add: AddCommandConfig = field(default_factory=lambda: AddCommandConfig())
    """The nested section for settings related to the `add` command."""
    delete: DeleteCommandConfig = field(default_factory=lambda: DeleteCommandConfig())
    """The nested section for settings related to the `delete` command."""
    edit: EditCommandConfig = field(default_factory=lambda: EditCommandConfig())
    """The nested section for settings related to the `edit` command."""
    import_: ImportCommandConfig = field(default_factory=lambda: ImportCommandConfig())
    """The nested section for settings related to the `import` command. Note the trailing underscore
    of its name, since this attribute would otherwise clash with the builtin `import` keyword."""
    list_: ListCommandConfig = field(default_factory=lambda: ListCommandConfig())
    """The nested section for settings related to the `list` command. Note the trailing underscore
    of its name, since this attribute would otherwise clash with the builtin `list` keyword."""
    modify: ModifyCommandConfig = field(default_factory=lambda: ModifyCommandConfig())
    """The nested section for settings related to the `modify` command."""
    open: OpenCommandConfig = field(default_factory=lambda: OpenCommandConfig())
    """The nested section for settings related to the `open` command."""
    search: SearchCommandConfig = field(default_factory=lambda: SearchCommandConfig())
    """The nested section for settings related to the `search` command."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS configuration section.")
        self.add.validate()
        self.delete.validate()
        self.edit.validate()
        self.import_.validate()
        self.list_.validate()
        self.modify.validate()
        self.open.validate()
        self.search.validate()


class LabelSuffix(Enum):
    """Suffixes to disambiguate `cobib.database.Entry` labels."""

    ALPHA = lambda count: chr(96 + count)  # pylint: disable=unnecessary-lambda-assignment
    CAPTIAL = lambda count: chr(64 + count)  # pylint: disable=unnecessary-lambda-assignment
    # pylint: disable=unnecessary-lambda,unnecessary-lambda-assignment
    NUMERIC = lambda count: str(count)


@dataclass
class DatabaseFormatConfig(_ConfigBase):
    """The `config.database.format` section."""

    label_default: str = "{label}"
    """Specifies a default label format which will be used for database entry keys. The format of
    this option follows the f-string like formatting of modifications (see also the documentation
    of the `cobib.commands.modify.ModifyCommand`). The default configuration value leaves the label
    unchanged compared to the metadata provided by the source from which the entry gets added. A
    more useful example is `"{author.split(' and ')[0].split()[-1]}{year}"` which takes the surname
    of the first author and immediately appends the publication year."""
    label_suffix: tuple[str, LabelSuffix] = field(default_factory=lambda: ("_", LabelSuffix.ALPHA))
    """Specifies the suffix format which is used to disambiguate labels if a conflict would occur.
    This option takes a tuple of length 2, where the first entry is the string separating the
    proposed label from the enumerator and the second one is one of the enumerators provided by the
    `LabelSuffix` object. The available enumerators are:
        - ALPHA: a, b, ...
        - CAPTIAL: A, B, ...
        - NUMERIC: 1, 2, ...
    """
    suppress_latex_warnings: bool = True
    """Specifies whether latex warnings should not be ignored during the escaping of special
    characters. This is a simple option which gets passed on to the internally used `pylatexenc`
    library."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE.FORMAT configuration section.")
        self._assert(
            isinstance(self.label_default, str),
            "config.database.format.label_default should be a string.",
        )
        self._assert(
            isinstance(self.label_suffix, tuple) and len(self.label_suffix) == 2,
            "config.database.format.label_suffix should be a tuple of length 2.",
        )
        self._assert(
            isinstance(self.label_suffix[0], str),
            "The first entry of config.database.format.label_suffix should be a string.",
        )
        self._assert(
            callable(self.label_suffix[1]),
            "The first entry of config.database.format.label_suffix should be a function.",
        )
        self._assert(
            isinstance(self.suppress_latex_warnings, bool),
            "config.database.format.suppress_latex_warnings should be a boolean.",
        )


@dataclass
class EntryListSeparatorConfig(_ConfigBase):
    """The `config.database.stringify.list_separator` section.

    These settings configure how list-based `cobib.database.Entry` fields are transformed into
    strings when converting to the BibTeX format. Each of these fields will be joined by the
    respective values.
    """

    file: str = ", "
    """Specifies the string used to join the list of files into a single string representation."""
    tags: str = ", "
    """Specifies the string used to join the list of tags into a single string representation."""
    url: str = ", "
    """Specifies the string used to join the list of URLs into a single string representation."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE.STRINGIFY.LIST_SEPARATOR configuration section.")
        self._assert(
            isinstance(self.file, str),
            "config.database.stringify.list_separator.file should be a string.",
        )
        self._assert(
            isinstance(self.tags, str),
            "config.database.stringify.list_separator.tags should be a string.",
        )
        self._assert(
            isinstance(self.url, str),
            "config.database.stringify.list_separator.url should be a string.",
        )


@dataclass
class EntryStringifyConfig(_ConfigBase):
    """The `config.database.stringify` section."""

    list_separator: EntryListSeparatorConfig = field(
        default_factory=lambda: EntryListSeparatorConfig()
    )
    """The nested section for list separator values."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE.STRINGIFY configuration section.")
        self.list_separator.validate()


@dataclass
class DatabaseConfig(_ConfigBase):
    """The `config.database` section."""

    file: str | Path = "~/.local/share/cobib/literature.yaml"
    """Specifies the path to the database YAML file. You can use `~` to represent your `$HOME`
    directory."""
    format: DatabaseFormatConfig = field(default_factory=lambda: DatabaseFormatConfig())
    """The nested section for database formatting settings."""
    git: bool = False
    """coBib can integrate with `git` in order to automatically track the history of the database.
    However, by default, this option is disabled. In order to make use of this, enable this setting
    and initialize your database with `cobib init --git`.

    .. warning::
       Before enabling this setting you must ensure that you have set up git properly by setting
       your name and email address."""
    stringify: EntryStringifyConfig = field(default_factory=lambda: EntryStringifyConfig())
    """The nested section for database string-formatting settings."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE configuration section.")
        self._assert(isinstance(self.file, str), "config.database.file should be a string.")
        self._assert(isinstance(self.git, bool), "config.database.git should be a boolean.")
        self.format.validate()
        self.stringify.validate()


@dataclass
class BibtexParserConfig(_ConfigBase):
    """The `config.parsers.bibtex` section."""

    ignore_non_standard_types: bool = False
    """Specifies whether the BibTeX-parser should ignore non-standard BibTeX entry types."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the PARSERS.BIBTEX configuration section.")
        self._assert(
            isinstance(self.ignore_non_standard_types, bool),
            "config.parsers.bibtex.ignore_non_standard_types should be a boolean.",
        )


@dataclass
class YAMLParserConfig(_ConfigBase):
    """The `config.parsers.yaml` section."""

    use_c_lib_yaml: bool = True
    """Specifies whether the C-based implementation of the YAML parser (called `LibYAML`) shall be
    used, *significantly* increasing the performance of the parsing.

    .. note::
       This requires manual installation of the C-based parser:
       https://yaml.readthedocs.io/en/latest/install.html#optional-requirements
    """

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the PARSERS.YAML configuration section.")
        self._assert(
            isinstance(self.use_c_lib_yaml, bool),
            "config.parsers.yaml.use_c_lib_yaml should be a boolean.",
        )


@dataclass
class ParserConfig(_ConfigBase):
    """The `config.parsers` section."""

    bibtex: BibtexParserConfig = field(default_factory=lambda: BibtexParserConfig())
    """The nested section for the BibTeX parser settings."""
    yaml: YAMLParserConfig = field(default_factory=lambda: YAMLParserConfig())
    """The nested section for the YAML parser settings."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the PARSERS configuration section.")
        self.bibtex.validate()
        self.yaml.validate()


@dataclass
class FileDownloaderConfig(_ConfigBase):
    """The `config.utils.file_downloader` section."""

    default_location: str = "~/.local/share/cobib"
    """Specifies the default download location for associated files."""
    url_map: dict[str, str] = field(default_factory=dict)
    """Permits providing rules to map from a journal's landing page URL to its PDF URL. To do so,
    insert an entry into this dictionary, with a regex-pattern matching the journal's landing page
    URL and a value being the PDF URL. E.g.:

    ```python
    config.utils.file_downloader.url_map[
        r"(.+)://aip.scitation.org/doi/([^/]+)"
    ] = r"\1://aip.scitation.org/doi/pdf/\2"

    config.utils.file_downloader.url_map[
        r"(.+)://quantum-journal.org/papers/([^/]+)"
    ] = r"\1://quantum-journal.org/papers/\2/pdf/"
    ```

    Make sure to use raw Python strings to ensure proper backslash-escaping."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the UTILS.FILE_DOWNLOADER configuration section.")
        self._assert(
            isinstance(self.default_location, str),
            "config.utils.file_downloader.default_location should be a string.",
        )
        self._assert(
            isinstance(self.url_map, dict),
            "config.utils.file_downloader.url_map should be a dict.",
        )
        for pattern, repl in self.url_map.items():
            self._assert(
                isinstance(pattern, str) and isinstance(repl, str),
                "config.utils.file_downloader.url_map should be a dict[str, str].",
            )


@dataclass
class UtilsConfig(_ConfigBase):
    """The `config.utils` section."""

    file_downloader: FileDownloaderConfig = field(default_factory=lambda: FileDownloaderConfig())
    """The nested section for the `cobib.utils.FileDownloader` utils settings."""
    journal_abbreviations: list[tuple[str, str]] = field(default_factory=list)
    """Permits providing a list of journal abbreviations. This list should be formatted as tuples of
    the form: `(full journal name, abbreviation)`. The abbreviation should include any necessary
    punctuation which can be excluded upon export (see also `cobib export --help`)."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the UTILS configuration section.")
        self.file_downloader.validate()
        self._assert(
            isinstance(self.journal_abbreviations, list),
            "config.utils.journal_abbreviations should be a list.",
        )
        for abbrev in self.journal_abbreviations:
            self._assert(
                isinstance(abbrev, tuple),
                "config.utils.journal_abbreviations should be a list of tuples.",
            )


@dataclass
class Config(_ConfigBase):
    """The `config` dataclass."""

    logging: LoggingConfig = field(default_factory=lambda: LoggingConfig())
    """The nested section for the logging settings."""
    commands: CommandConfig = field(default_factory=lambda: CommandConfig())
    """The nested section for the commands settings."""
    database: DatabaseConfig = field(default_factory=lambda: DatabaseConfig())
    """The nested section for the database settings."""
    events: dict["Event", list[Callable]] = field(default_factory=dict)  # type: ignore[type-arg]
    """It is possible to register hooks on various events. Although this can be done manually using
    this dictionary, it is preferred to use the function-decorators like so:
    To subscribe to a certain event do something similar to the following:

    ```python
    from os import system
    from cobib.config import Event
    from cobib.commands import InitCommand

    @Event.PostInitCommand.subscribe
    def add_remote(cmd: InitCommand) -> None:
        system(f"git -C {cmd.root} remote add origin https://github.com/user/repo")
    ```

    Note, that the typing is required for the config validation to pass!
    For more information refer to the `cobib.config.Event` documentation.
    """
    parsers: ParserConfig = field(default_factory=lambda: ParserConfig())
    """The nested section for the parsers settings."""
    utils: UtilsConfig = field(default_factory=lambda: UtilsConfig())
    """The nested section for the utils settings."""

    # pylint: disable=invalid-name
    XDG_CONFIG_FILE: str | Path = field(
        default="~/.config/cobib/config.py", init=False, repr=False, compare=False
    )
    """The XDG-based standard configuration location."""

    @override
    def validate(self) -> None:
        LOGGER.info("Validating the runtime configuration.")
        self.logging.validate()
        self.commands.validate()
        self.database.validate()
        self.parsers.validate()
        self.utils.validate()

        LOGGER.debug("Validating the EVENTS configuration section.")
        self._assert(isinstance(self.events, dict), "config.events should be a dict.")
        for event in self.events:
            self._assert(
                event.validate(),
                f"config.events.{event} did not pass its validation check.",
            )

    @staticmethod
    def load(configpath: Optional[Union[str, Path, TextIO, io.TextIOWrapper]] = None) -> None:
        """Loads another configuration object at runtime.

        WARNING: The new Python-like configuration allows essentially arbitrary Python code so it is
        the user's responsibility to treat this with care!

        Args:
            configpath: the path to the configuration.
        """
        LOGGER.info(configpath)
        if configpath is not None:
            if isinstance(configpath, (TextIO, io.TextIOWrapper)):
                configpath.close()
                configpath = configpath.name
        elif "COBIB_CONFIG" in os.environ:
            configpath_env = os.environ["COBIB_CONFIG"]
            if configpath_env.lower() in ("", "0", "f", "false", "nil", "none"):
                LOGGER.info(
                    "Skipping configuration loading because negative COBIB_CONFIG environment "
                    "variable was detected."
                )
                return
            configpath = RelPath(configpath_env).path
        elif Config.XDG_CONFIG_FILE and RelPath(Config.XDG_CONFIG_FILE).exists():
            configpath = RelPath(Config.XDG_CONFIG_FILE).path
        else:  # pragma: no cover
            return  # pragma: no cover
        LOGGER.info("Loading configuration from default location: %s", configpath)

        spec = importlib.util.spec_from_file_location("config", configpath)
        if spec is None:
            LOGGER.error(
                "The config at %s could not be interpreted as a Python module.", configpath
            )
            sys.exit(1)
        else:
            cfg = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(cfg)  # type: ignore

        try:
            # validate config
            config.validate()
        except RuntimeError as exc:
            LOGGER.error(exc)
            sys.exit(1)


config = Config()
"""This is the runtime configuration object. It is exposed on the module level via:
```python
from cobib.config import config
```
"""
config.defaults()



"""coBib - the Console Bibliography.

.. include:: ../../README.md
"""

import subprocess
from pathlib import Path

__version__ = "4.1.0"

if (Path(__file__).parent.parent.parent / ".git").exists():
    # if installed from source, append HEAD commit SHA to version info as metadata
    with subprocess.Popen(["git", "rev-parse", "HEAD"], stdout=subprocess.PIPE) as proc:
        git_revision, _ = proc.communicate()
    __version__ += "+" + git_revision.decode("utf-8")[:7]



"""coBib's help screen.

This screen renders a help information with the current key bindings and action descriptions.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from rich.table import Table
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Static
from typing_extensions import override


class HelpScreen(ModalScreen[None]):
    """coBib's help screen."""

    BINDINGS = [
        ("question_mark", "toggle_help", "Help"),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | ? | Toggles the help screen. |
    """

    DEFAULT_CSS = """
        HelpScreen {
            align: center middle;
        }

        #help {
            padding: 1 2;
            width: auto;
            height: auto;
            background: $surface;
        }
    """

    HELP_DESCRIPTIONS = [
        ("q", "Quit's coBib"),
        ("question_mark", "Toggles the help screen"),
        ("underscore", "Toggles between the horizontal and vertical layout"),
        ("space", "Toggles folding of a search result"),
        ("colon", "Starts the prompt for an arbitrary coBib command"),
        ("v", "Selects the current entry"),
        ("slash", "Searches the database for the provided string"),
        ("a", "Prompts for a new entry to be added to the database"),
        ("d", "Deletes the current (or selected) entries"),
        ("e", "Edits the current entry"),
        ("f", "Allows filtering the table using `++/--` keywords"),
        ("i", "Imports entries from another source"),
        ("m", "Prompts for a modification (respects selection)"),
        ("o", "Opens the current (or selected) entries"),
        ("r", "Redoes the last undone change. Requires git-tracking!"),
        ("s", "Prompts for the field to sort by (use -r to list in reverse)"),
        ("u", "Undes the last change. Requires git-tracking!"),
        ("x", "Exports the current (or selected) entries"),
        ("j", "Moves one row down"),
        ("k", "Moves one row up"),
        ("h", "Moves to the left"),
        ("l", "Moves to the right"),
    ]
    """The key binding help descriptions."""

    @override
    def compose(self) -> ComposeResult:
        help_table = Table(title="coBib TUI Help", caption="Close this help by pressing '?'")
        help_table.add_column("Key")
        help_table.add_column("Description")
        for key, description in self.HELP_DESCRIPTIONS:
            help_table.add_row(self.app.get_key_display(key), description)
        static = Static(help_table, id="help")
        static.styles.content_align = ("center", "middle")
        yield static

    def action_toggle_help(self) -> None:
        """Toggles the help information.

        Since this is the action of the `HelpScreen`, it simply pops the screen.
        """
        self.app.pop_screen()



"""coBib's popup panel widget.

This widget is merely a container for placing all the `cobib.ui.components.popup.Popup` instances.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from textual.containers import Container


class PopupPanel(Container, can_focus=False, can_focus_children=False):
    """coBib's popup panel widget."""

    DEFAULT_CSS = """
        PopupPanel {
            layer: popup;
            layout: vertical;
            align: center bottom;
            width: 100%;
            height: auto;
            offset-y: -1;
        }
    """



"""Tests for coBib's DOIParser."""
# pylint: disable=unused-argument

import logging
from typing import Dict, Optional

import pytest
import requests

from cobib.config import Event
from cobib.database import Entry
from cobib.parsers import DOIParser

from .parser_test import ParserTest


def assert_default_test_entry(entry: Entry) -> None:
    """Asserts that the passed entry is the default testing entry.

    Args:
        entry: the entry to assert.
    """
    reference = ParserTest.EXAMPLE_ENTRY_DICT.copy()
    # In this specific case the bib file provided by this DOI includes additional (yet
    # unnecessary) brackets in the escaped special characters of the author field. Thus, we
    # correct for this inconsistency manually before asserting the equality.
    reference["author"] = str(reference["author"]).replace("'a", "'{a}")
    reference["_download"] = "https://pubs.acs.org/doi/10.1021/acs.chemrev.8b00803"
    assert entry.data == reference


class TestDOIParser(ParserTest):
    """Tests for coBib's DOIParser."""

    @pytest.mark.parametrize(
        "query", ["10.1021/acs.chemrev.8b00803", "https://doi.org/10.1021/acs.chemrev.8b00803"]
    )
    def test_from_doi(self, query: str, caplog: pytest.LogCaptureFixture) -> None:
        """Test parsing from DOI.

        Args:
            query: the DOI or URL which to query.
            caplog: the built-in pytest fixture.
        """
        entries = DOIParser().parse(query)

        if (
            "cobib.parsers.doi",
            logging.ERROR,
            "An Exception occurred while trying to query the DOI: 10.1021/acs.chemrev.8b00803.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    def test_invalid_doi(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test parsing an invalid DOI.

        Args:
            caplog: the built-in pytest fixture.
        """
        entries = DOIParser().parse("1812.09976")
        assert not entries
        assert entries == {}  # pylint: disable=C1803

        assert (
            "cobib.parsers.doi",
            logging.WARNING,
            "'1812.09976' is not a valid DOI.",
        ) in caplog.record_tuples

    def test_catching_api_error(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test catching API error.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """

        def raise_exception(*args, **kwargs):  # type: ignore
            """Mock function to raise an Exception."""
            raise requests.exceptions.RequestException()

        monkeypatch.setattr(requests, "get", raise_exception)
        DOIParser().parse("10.1021/acs.chemrev.8b00803")

        assert (
            "cobib.parsers.doi",
            logging.ERROR,
            "An Exception occurred while trying to query the DOI: 10.1021/acs.chemrev.8b00803.",
        ) in caplog.record_tuples

    def test_dump(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test dumping.

        Args:
            caplog: the built-in pytest fixture.
        """
        entry = Entry("dummy", {"ENTRYTYPE": "unpublished"})
        DOIParser().dump(entry)

        assert (
            "cobib.parsers.doi",
            logging.ERROR,
            "Cannot dump an entry as a DOI.",
        ) in caplog.record_tuples

    def test_event_pre_doi_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PreDOIParse event."""

        @Event.PreDOIParse.subscribe
        def hook(string: str) -> Optional[str]:
            return "10.1021/acs.chemrev.8b00803"

        assert Event.PreDOIParse.validate()

        entries = DOIParser().parse("Hello world!")
        if (
            "cobib.parsers.doi",
            logging.ERROR,
            "An Exception occurred while trying to query the DOI: 10.1021/acs.chemrev.8b00803.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    def test_event_post_doi_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PostDOIParse event."""

        @Event.PostDOIParse.subscribe
        def hook(bib: Dict[str, Entry]) -> None:
            bib["Cao_2019"].data["test"] = "dummy"

        assert Event.PostDOIParse.validate()

        entries = DOIParser().parse("10.1021/acs.chemrev.8b00803")
        if (
            "cobib.parsers.doi",
            logging.ERROR,
            "An Exception occurred while trying to query the DOI: 10.1021/acs.chemrev.8b00803.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert entry.data["test"] == "dummy"



"""Tests for coBib's ModifyCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import contextlib
import tempfile
from io import StringIO
from typing import TYPE_CHECKING, Any, List, Type

import pytest
from typing_extensions import override

from cobib.commands import ModifyCommand
from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestModifyCommand(CommandTest):
    """Tests for coBib's ModifyCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return ModifyCommand

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["modification", "filters", "selection"],
        [
            ["tags:test", ["einstein"], True],
            ["tags:test", ["++label", "einstein"], False],
        ],
    )
    @pytest.mark.parametrize("add", [False, True])
    @pytest.mark.parametrize("dry", [False, True])
    def test_command(
        self,
        setup: Any,
        modification: str,
        filters: List[str],
        selection: bool,
        add: bool,
        dry: bool,
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            modification: the modification string to apply.
            filters: the filter arguments for the command.
            selection: whether the filters are of `selection` type.
            add: whether to use append mode.
            dry: whether to run in dry mode.
        """
        git = setup.get("git", False)

        # modify some data
        args = [modification, "--"] + filters
        if selection:
            args = ["-s"] + args

        expected = ["test"]

        if add:
            # first insert something to add to
            ModifyCommand("tags:dummy", "++label", "einstein").execute()
            args = ["-a"] + args
            expected = ["dummy"] + expected

        if dry:
            args.insert(0, "--dry")

        ModifyCommand(*args).execute()

        if dry:
            if add:
                assert Database()["einstein"].data["tags"] == ["dummy"]
            else:
                assert "tags" not in Database()["einstein"].data.keys()
        else:
            assert Database()["einstein"].data["tags"] == expected

        if git:
            try:
                # assert the git commit message
                self.assert_git_commit_message(
                    "modify",
                    {
                        "modification": modification.split(":"),
                        "dry": False,
                        "add": add,
                        "preserve_files": None,
                        "selection": selection,
                        "filter": filters,
                    },
                )
            except AssertionError:
                assert dry

    @pytest.mark.parametrize(
        ["modification", "expected"],
        [
            ["author: and Knuth", "Albert Einstein and Knuth"],
            ["dummy:test", "test"],
            ["number:2", 12],
            ["number:a", "10a"],
        ],
    )
    def test_add_mode(self, setup: Any, modification: str, expected: Any) -> None:
        """Test more cases of the add mode.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            modification: the modification string to apply.
            expected: the expected final `Entry` field.
        """
        # modify some data
        args = ["-a", modification, "++label", "einstein"]

        field, _ = modification.split(":")

        ModifyCommand(*args).execute()

        assert Database()["einstein"].data[field] == expected

    @pytest.mark.parametrize(
        ["modification", "expected"],
        [
            ["pages:{pages.replace('--', '-')}", "891-921"],
            ["year:{year+10}", 1915],
            ["label:{author.split()[1]}{year}", "Einstein1905"],
            ["string:{'à' !a}", "'\\xe0'"],
            ["number:{1.2345:.2}", "1.2"],
            ["dummy:{dummy}", ""],
        ],
    )
    def test_f_string_interpretation(self, setup: Any, modification: str, expected: Any) -> None:
        """Test f-string interpretation.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            modification: the modification string to apply.
            expected: the expected final `Entry` field.
        """
        # modify some data
        args = [modification, "++label", "einstein"]

        field, *_ = modification.split(":")

        ModifyCommand(*args).execute()

        if field != "label":
            assert Database()["einstein"].data[field] == expected
        else:
            assert "eistein" not in Database().keys()
            assert expected in Database().keys()
            assert Database()[expected].label == expected

    @pytest.mark.parametrize("preserve_files", [None, True, False])
    @pytest.mark.parametrize("config_overwrite", [True, False])
    def test_rename_associated_file(
        self, setup: Any, preserve_files: bool, config_overwrite: bool
    ) -> None:
        """Test removing associated files.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            preserve_files: argument to `ModifyCommand`.
            config_overwrite: what to overwrite `config.commands.modify.preserve_files` with.
        """
        config.commands.modify.preserve_files = config_overwrite

        should_preserve = config_overwrite
        if preserve_files is not None:
            should_preserve = preserve_files

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = RelPath(tmpdirname + "/knuthwebsite.pdf")
            open(path.path, "w", encoding="utf-8").close()  # pylint: disable=consider-using-with

            Database()["knuthwebsite"].file = str(path)

            args = ["label:dummy", "-s", "--", "knuthwebsite"]
            if preserve_files is not None:
                args.insert(2, f"--{'' if preserve_files else 'no-'}preserve-files")
            ModifyCommand(*args).execute()
            assert "dummy" in Database().keys()

            target = RelPath(tmpdirname + "/dummy.pdf")
            if should_preserve:
                assert path.path.exists()
            else:
                assert target.path.exists()

    def test_warning_missing_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing label.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        # Note: when using a filter, no non-existent label can occur
        args = ["-s", "tags:test", "--", "dummy"]
        ModifyCommand(*args).execute()
        assert (
            "cobib.commands.modify",
            30,
            "No entry with the label 'dummy' could be found.",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["args"],
        [
            [["-s", "tags:test", "--", "einstein"]],
            [["tags:test", "--", "++label", "einstein"]],
        ],
    )
    # other variants are already covered by test_command
    async def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, args: List[str]
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            args: additional arguments to pass to the command.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "modify"] + args)
        assert Database()["einstein"].data["tags"] == ["test"]

    def test_event_pre_modify_command(self, setup: Any) -> None:
        """Tests the PreModifyCommand event."""

        @Event.PreModifyCommand.subscribe
        def hook(command: ModifyCommand) -> None:
            command.largs.modification = ("number", "2")

        assert Event.PreModifyCommand.validate()

        ModifyCommand("-a", "number:3", "++label", "einstein").execute()

        assert Database()["einstein"].data["number"] == 12

    def test_event_post_modify_command(self, setup: Any) -> None:
        """Tests the PostModifyCommand event."""

        @Event.PostModifyCommand.subscribe
        def hook(command: ModifyCommand) -> None:
            print([entry.label for entry in command.modified_entries])

        assert Event.PostModifyCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as out:
            ModifyCommand("-a", "number:3", "++label", "einstein").execute()
            assert out.getvalue() == "['einstein']\n"



"""Tests for coBib's Journal abbreviations."""
# pylint: disable=protected-access

from typing import Any, Generator

import pytest

from cobib.config import config
from cobib.utils.journal_abbreviations import JournalAbbreviations


@pytest.fixture(autouse=True)
def setup() -> Generator[Any, None, None]:
    """Setup.

    This method is mostly used to reset the JournalAbbreviations to their original state.

    Yields:
        Access to the local fixture variables.
    """
    yield
    config.utils.journal_abbreviations = []
    JournalAbbreviations._abbreviations = {}
    JournalAbbreviations._fullwords = {}


def test_downloader_singleton() -> None:
    """Test the JournalAbbreviations is a Singleton."""
    j_a = JournalAbbreviations()
    j_a2 = JournalAbbreviations()
    assert j_a is j_a2


def test_load_abbreviations() -> None:
    """Test the JournalAbbreviations.load_abbreviations method."""
    config.utils.journal_abbreviations = [("Test Journal", "Test J.")]
    JournalAbbreviations.load_abbreviations()
    assert JournalAbbreviations._abbreviations == {"Test Journal": "Test J."}
    assert JournalAbbreviations._fullwords == {"Test J.": "Test Journal", "Test J": "Test Journal"}


def test_check_existence(caplog: pytest.LogCaptureFixture) -> None:
    """Test the JournalAbbreviations.check_existence method.

    Args:
        caplog: the built-in pytest fixture.
    """
    assert JournalAbbreviations.check_existence("Test Journal") is False
    for scope, level, message in caplog.record_tuples:
        if (
            scope == "cobib.utils.journal_abbreviations"
            and level == 30
            and "'Test Journal' was not found" in message
        ):
            break
    else:
        assert False, "Warning not raised upon missing journal!"
    caplog.clear()
    config.utils.journal_abbreviations = [("Test Journal", "Test J.")]
    assert JournalAbbreviations.check_existence("Test Journal")
    assert JournalAbbreviations.check_existence("Test J.")
    assert JournalAbbreviations.check_existence("Test J")


@pytest.mark.parametrize("dotless", [False, True])
def test_abbreviate(dotless: bool) -> None:
    """Test the JournalAbbreviations.abbreviate method.

    Args:
        dotless: whether to abbreviate with or without punctuation.
    """
    old_journal = "Test Journal"
    new_journal = JournalAbbreviations.abbreviate(old_journal, dotless=dotless)
    assert new_journal == old_journal
    config.utils.journal_abbreviations = [("Test Journal", "Test J.")]
    new_journal = JournalAbbreviations.abbreviate(old_journal, dotless=dotless)
    expected = "Test J" if dotless else "Test J."
    assert new_journal == expected


def test_elongate() -> None:
    """Test the JournalAbbreviations.elongate method."""
    old_journal = "Test J."
    new_journal = JournalAbbreviations.elongate(old_journal)
    assert new_journal == old_journal
    config.utils.journal_abbreviations = [("Test Journal", "Test J.")]
    new_journal = JournalAbbreviations.elongate(old_journal)
    assert new_journal == "Test Journal"
    old_journal = "Test Journal"
    new_journal = JournalAbbreviations.elongate(old_journal)
    assert new_journal == "Test Journal"



"""coBib parser test class."""

import pytest

from cobib.config import config

from .. import get_resource


class ParserTest:
    """The base class for coBib's parser test classes."""

    EXAMPLE_BIBTEX_FILE = get_resource("example_entry.bib")
    """Path to the example BibTeX file."""

    EXAMPLE_YAML_FILE = get_resource("example_entry.yaml")
    """Path to the example YAML file (matching the BibTeX file)."""

    EXAMPLE_ENTRY_DICT = {
        "ENTRYTYPE": "article",
        "author": "Yudong Cao and Jonathan Romero and Jonathan P. Olson and Matthias Degroote and "
        + "Peter D. Johnson and M{\\'a}ria Kieferov{\\'a} and Ian D. Kivlichan and Tim Menke and "
        + "Borja Peropadre and Nicolas P. D. Sawaya and Sukin Sim and Libor Veis and Al{\\'a}n "
        + "Aspuru-Guzik",
        "doi": "10.1021/acs.chemrev.8b00803",
        "journal": "Chemical Reviews",
        "month": "aug",
        "number": 19,
        "pages": "10856--10915",
        "publisher": "American Chemical Society ({ACS})",
        "title": "Quantum Chemistry in the Age of Quantum Computing",
        "url": ["https://doi.org/10.1021%2Facs.chemrev.8b00803"],
        "volume": 119,
        "year": 2019,
    }
    """The matching dictionary to the example files also included here."""

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        """Setup."""
        config.defaults()



"""coBib's Undo command.

This command can be used to undo the changes of the a previous command.
```
cobib undo
```
For obvious reasons, this will only undo commands which had an effect on the contents of the
database file.
Moreover, as a safety measure, this command will only undo those changes, which have been committed
by coBib *automatically*.
You can disable this be setting the `--force` argument which *always* undoes the last commit:
```
cobib undo --force
```

.. warning::
   This command is *only* available if coBib's git-integration has been enabled via
   `cobib.config.config.DatabaseConfig.git` *and* initialized properly (see `cobib.commands.init`).

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `u` key.
"""

from __future__ import annotations

import logging
import subprocess
import sys
from pathlib import Path
from typing import Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class UndoCommand(Command):
    """The Undo Command.

    This command can parse the following arguments:

        * `-f`, `--force`: if specified, this will also revert changes which have *not* been
          auto-committed by coBib.
    """

    name = "undo"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.root: Path
        """The path to the root of the git repository tracking the database."""

        self.sha: str
        """The git commit SHA which was reverted by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="undo", description="Undo subcommand parser.")
        parser.add_argument(
            "-f", "--force", action="store_true", help="allow undoing non auto-committed changes"
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        git_tracked = config.database.git
        if not git_tracked:
            msg = (
                "You must enable coBib's git-tracking in order to use the `Undo` command."
                "\nPlease refer to the man-page for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        file = RelPath(config.database.file).path
        self.root = file.parent
        if not (self.root / ".git").exists():
            msg = (
                "You have configured, but not initialized coBib's git-tracking."
                "\nPlease consult `cobib init --help` for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        LOGGER.debug("Starting Undo command.")

        Event.PreUndoCommand.fire(self)

        LOGGER.debug("Obtaining git log.")
        lines = subprocess.check_output(
            [
                "git",
                "--no-pager",
                "-C",
                f"{self.root}",
                "log",
                "--oneline",
                "--no-decorate",
                "--no-abbrev",
            ]
        )
        undone_shas = set()
        for commit in lines.decode().strip().split("\n"):
            LOGGER.debug("Processing commit %s", commit)
            self.sha, *message = commit.split()
            if message[0] == "Undo":
                # Store already undone commit sha
                LOGGER.debug("Storing undone commit sha: %s", message[-1])
                undone_shas.add(message[-1])
                continue
            if self.sha in undone_shas:
                LOGGER.info("Skipping %s as it was already undone", self.sha)
                continue
            if self.largs.force or (message[0] == "Auto-commit:" and message[-1] != "InitCommand"):
                # we undo a commit if and only if:
                #  - the `force` argument is specified OR
                #  - the commit is an `auto-committed` change which is NOT from `InitCommand`
                LOGGER.debug("Attempting to undo %s.", self.sha)
                commands = [
                    f"git -C {self.root} revert --no-commit {self.sha}",
                    f"git -C {self.root} commit --no-gpg-sign --quiet --message 'Undo {self.sha}'",
                ]
                with subprocess.Popen(
                    "; ".join(commands), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                ) as undo:
                    undo.communicate()
                    if undo.returncode != 0:
                        LOGGER.error(  # pragma: no cover
                            "Undo was unsuccessful. Please consult the logs and git history of your"
                            " database for more information."
                        )
                    else:
                        # update Database
                        Database().read()
                break
        else:
            msg = "Could not find a commit to undo. Please commit something first!"
            LOGGER.warning(msg)
            sys.exit(1)

        Event.PostUndoCommand.fire(self)



#!/usr/bin/env python3
"""coBib main body."""

import asyncio
import sys

from cobib.ui.cli import CLI
from cobib.ui.shell_helper import ShellHelper


async def main() -> None:
    """Main async executable.

    coBib's main function used to parse optional keyword arguments and subcommands.
    """
    if len(sys.argv) > 1 and any(a[0] == "_" for a in sys.argv):
        # shell helper function called
        ShellHelper().run()
    else:
        await CLI().run()


def _main() -> None:
    """The main method wrapping the async method with `asyncio.run`."""
    asyncio.run(main())


if __name__ == "__main__":
    _main()  # pragma: no cover



"""coBib's Redo command.

This command can be used to re-apply the changes *of a previously undone* command (see
`cobib.commands.undo`):
```
cobib redo
```
This command takes *no* additional arguments!

Note, that if you have not used `cobib undo` previously, this command will have no effect!

.. warning::
   This command is *only* available if coBib's git-integration has been enabled via
   `cobib.config.config.DatabaseConfig.git` *and* initialized properly (see `cobib.commands.init`).

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `r` key.
"""

from __future__ import annotations

import logging
import subprocess
import sys
from pathlib import Path
from typing import Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class RedoCommand(Command):
    """The Redo Command.

    This command does not parse any additional arguments.
    """

    name = "redo"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.root: Path
        """The path to the root of the git repository tracking the database."""

        self.sha: str
        """The git commit SHA which was reverted by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="redo", description="Redo subcommand parser.")
        cls.argparser = parser

    @override
    def execute(self) -> None:
        git_tracked = config.database.git
        if not git_tracked:
            msg = (
                "You must enable coBib's git-tracking in order to use the `Redo` command."
                "\nPlease refer to the man-page for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        file = RelPath(config.database.file).path
        self.root = file.parent
        if not (self.root / ".git").exists():
            msg = (
                "You have configured, but not initialized coBib's git-tracking."
                "\nPlease consult `cobib init --help` for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        LOGGER.debug("Starting Redo command.")

        Event.PreRedoCommand.fire(self)

        LOGGER.debug("Obtaining git log.")
        lines = subprocess.check_output(
            [
                "git",
                "--no-pager",
                "-C",
                f"{self.root}",
                "log",
                "--oneline",
                "--no-decorate",
                "--no-abbrev",
            ]
        )
        redone_shas = set()
        for commit in lines.decode().strip().split("\n"):
            LOGGER.debug("Processing commit %s", commit)
            self.sha, *message = commit.split()
            if message[0] == "Redo":
                # Store already redone commit sha
                LOGGER.debug("Storing redone commit sha: %s", message[-1])
                redone_shas.add(message[-1])
                continue
            if self.sha in redone_shas:
                LOGGER.info("Skipping %s as it was already redone", self.sha)
                continue
            if message[0] == "Undo":
                LOGGER.debug("Attempting to redo %s.", self.sha)
                commands = [
                    f"git -C {self.root} revert --no-commit {self.sha}",
                    f"git -C {self.root} commit --no-gpg-sign --quiet --message 'Redo {self.sha}'",
                ]
                with subprocess.Popen(
                    "; ".join(commands), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                ) as redo:
                    redo.communicate()
                    if redo.returncode != 0:
                        LOGGER.error(  # pragma: no cover
                            "Redo was unsuccessful. Please consult the logs and git history of your"
                            " database for more information."
                        )
                    else:
                        # update Database
                        Database().read()
                break
        else:
            msg = "Could not find a commit to redo. You must have undone something first!"
            LOGGER.warning(msg)
            sys.exit(1)

        Event.PostRedoCommand.fire(self)



"""coBib's popup widget.

This widget renders as a simple popup. For example, it is used to display logging information.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

import logging
from typing import Any

from rich.console import RenderableType
from textual.widgets import Label


class Popup(Label, can_focus=False, can_focus_children=False):
    """coBib's popup widget."""

    DEFAULT_CSS = """
        Popup {
            text-style: bold;
            min-height: 1;
            width: 100%;
            color: auto;
        }
    """

    def __init__(
        self,
        renderable: RenderableType,
        *args: Any,
        level: int = logging.INFO,
        timer: float | None = 5.0,
        **kwargs: Any,
    ) -> None:
        """Initializes the popup.

        Args:
            renderable: the contents of the popup.
            *args: any further positional arguments for the underlying `Label` class.
            level: the `logging` level which this popup references. The popup's back- and foreground
                colors are styled according to this setting. Setting this to `0` will use the
                default styling.
            timer: an optional number of seconds after which to automatically remove the popup.
            **kwargs: any further keyword arguments for the underlying `Label` class.
        """
        super().__init__(renderable, *args, **kwargs)
        if level >= logging.CRITICAL:
            self.styles.background = "red"
            self.styles.color = "yellow"
        elif level >= logging.ERROR:
            self.styles.background = "red"
        elif level >= logging.WARNING:
            self.styles.background = "yellow"
        elif level >= logging.INFO:
            self.styles.background = "green"
        elif level >= logging.DEBUG:
            self.styles.background = "blue"
        if timer is not None:
            self.set_timer(timer, self.remove)



"""Tests for coBib's Event validation."""

from typing import Any, Generator

import pytest

from cobib.config import Event, config

from .. import get_resource


@pytest.fixture(autouse=True)
def setup() -> Generator[Any, None, None]:
    """Setup debugging configuration.

    Yields:
        Access to the local fixture variables.
    """
    config.load(get_resource("debug.py"))
    yield setup
    config.defaults()


@pytest.mark.parametrize(
    "event",
    [
        Event.PreAddCommand,
        Event.PostAddCommand,
        Event.PreDeleteCommand,
        Event.PostDeleteCommand,
        Event.PreEditCommand,
        Event.PostEditCommand,
        Event.PreExportCommand,
        Event.PostExportCommand,
        Event.PreImportCommand,
        Event.PostImportCommand,
        Event.PreInitCommand,
        Event.PostInitCommand,
        Event.PreListCommand,
        Event.PostListCommand,
        Event.PreModifyCommand,
        Event.PostModifyCommand,
        Event.PreOpenCommand,
        Event.PostOpenCommand,
        Event.PreRedoCommand,
        Event.PostRedoCommand,
        Event.PreSearchCommand,
        Event.PostSearchCommand,
        Event.PreShowCommand,
        Event.PostShowCommand,
        Event.PreUndoCommand,
        Event.PostUndoCommand,
        Event.PreBibtexParse,
        Event.PostBibtexParse,
        Event.PreBibtexDump,
        Event.PostBibtexDump,
        Event.PreYAMLParse,
        Event.PostYAMLParse,
        Event.PreYAMLDump,
        Event.PostYAMLDump,
        Event.PreArxivParse,
        Event.PostArxivParse,
        Event.PreDOIParse,
        Event.PostDOIParse,
        Event.PreISBNParse,
        Event.PostISBNParse,
        Event.PreURLParse,
        Event.PostURLParse,
        Event.PreZoteroImport,
        Event.PostZoteroImport,
        Event.PreFileDownload,
        Event.PostFileDownload,
        Event.PreGitCommit,
        Event.PostGitCommit,
    ],
)
def test_config_validation(event: Event) -> None:
    """Test the config validation of faulty hook subscription functions.

    Args:
        event: the coBib Event which to subscribe to.
    """

    @event.subscribe
    def faulty_hook() -> None:
        return

    with pytest.raises(RuntimeError) as exc_info:
        config.validate()

    assert (
        str(exc_info.value)
        == f"config.events.Event.{event.name} did not pass its validation check."
    )



"""Tests for coBib's logging helper functions."""

import logging
import tempfile
from pathlib import Path
from typing import Optional

import pytest
from rich.console import Group
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text

from cobib.utils.logging import get_file_handler, get_stream_handler, print_changelog

from .. import get_resource


def test_get_stream_handler() -> None:
    """Test stream logging configuration."""
    handler = get_stream_handler()
    assert handler.level == 30
    assert isinstance(handler, logging.StreamHandler)


def test_get_file_handler() -> None:
    """Test file logging configuration."""
    handler = get_file_handler("INFO")
    assert handler.level == 20
    assert isinstance(handler, logging.FileHandler)


@pytest.mark.parametrize("cached_version", [None, False, "", "0.1", "0.2"])
def test_print_changelog(
    cached_version: Optional[str],
) -> None:
    """Test printing of the latest changelog section.

    Args:
        cached_version: the cached version to test against.
    """
    if cached_version is None:
        panel = print_changelog("0.2", None)
        assert panel is None
    else:
        with tempfile.NamedTemporaryFile("w") as file:
            path = Path(file.name)
            if not cached_version:
                file.close()
            else:
                file.write(cached_version)
                file.flush()
            panel = print_changelog("0.2", str(path))
        # ensure we do not leave a temporary file behind
        if path.exists():
            path.unlink()
        if cached_version == "0.2":
            assert panel is None
        else:
            assert isinstance(panel, Panel)
            assert isinstance(panel.renderable, Group)
            assert isinstance(panel.renderable.renderables[0], Text)
            assert isinstance(panel.renderable.renderables[1], Markdown)
            with open(
                get_resource("expected_changelog_printing.md", "utils"), "r", encoding="utf-8"
            ) as expected:
                expected_lines = expected.read().splitlines()
                for true, exp in zip(
                    panel.renderable.renderables[1].markup.splitlines(), expected_lines
                ):
                    assert true == exp


def test_safe_cache_access() -> None:
    """Regression test against #94."""
    # create an empty temporary directory
    tmp_dir = tempfile.mkdtemp()
    # use a path which surely does not exist
    tmp_cache_file = Path(tmp_dir) / "cache/version"
    try:
        # try to read the version from a file whose parent directory does not exist
        print_changelog("0.2", str(tmp_cache_file))
        # if the test regresses, this will fail with a `FileNotFoundError`
    finally:
        if tmp_cache_file.exists():
            tmp_cache_file.unlink()
        if tmp_cache_file.parent.exists():
            tmp_cache_file.parent.rmdir()
        Path(tmp_dir).rmdir()



"""coBib's prompt widget.

This widget implements rich's
[`PromptBase`](https://rich.readthedocs.io/en/stable/reference/prompt.html#rich.prompt.PromptBase).

It leverages coBib's `cobib.ui.components.input_screen.InputScreen` for handling user input.

This implementation permits the seamless integration of interactive user prompts during command
execution with an interactive TUI based on textual.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, TextIO, cast

from rich.prompt import InvalidResponse, PromptBase
from rich.text import Text, TextType
from textual.widgets import Input, Static
from typing_extensions import override

from .input_screen import InputScreen

if TYPE_CHECKING:
    from ..tui import TUI


class Prompt(PromptBase[str]):
    """coBib's prompt widget."""

    console: TUI  # type: ignore[assignment]
    """The running TUI instance. This overloads the meaning of `console` in rich's understanding."""

    # pylint: disable=invalid-overridden-method
    @override
    async def __call__(  # type: ignore[override]
        self, *, default: Any = ..., stream: TextIO | None = None
    ) -> str:
        while True:
            self.pre_prompt()
            prompt = self.make_prompt(default)
            value = await self.get_input(self.console, prompt, self.password, stream=stream)
            if value == "" and default != ...:
                return str(default)
            try:
                return_value = self.process_response(value)
            except InvalidResponse as error:
                self.on_validate_error(value, error)
                continue
            else:
                return return_value

    # pylint: disable=invalid-overridden-method
    @override
    @classmethod
    async def get_input(  # type: ignore[override]
        cls, console: TUI, prompt: TextType, password: bool, stream: TextIO | None = None
    ) -> str:
        dismiss_event = asyncio.Event()
        value = ""

        def _catch_dismissed_value(val: str) -> None:
            nonlocal value
            value = val
            dismiss_event.set()

        popup = Static(prompt + "\n", id="prompt")

        inp_screen = cast(InputScreen, console.get_screen("input"))
        inp_screen.escape_enabled = False

        if inp_screen.is_current:
            # We need to pop the screen first to ensure that we can register our result callback.
            # The reason it might be the current screen already is if the pre_prompt has pushed it.
            console.pop_screen()

        await_mount = console.push_screen("input", _catch_dismissed_value)
        inp_screen.mount(popup, before=-1)
        await await_mount

        inp_screen.query_one(Input).value = ""
        await dismiss_event.wait()
        return value

    @override
    def on_validate_error(self, value: str, error: InvalidResponse) -> None:
        _id = "help" if value == "help" else "error"
        if isinstance(error.message, Text):
            message = Static(error.message, id=_id)
        else:
            message = Static(Text.from_markup(error.message), id=_id)

        self.console.get_screen("input").mount(message, before=-1)



"""coBib's Journal abbreviations."""

from __future__ import annotations

import logging
from typing import Callable, Dict, Optional

from cobib.config import config

LOGGER = logging.getLogger(__name__)


class JournalAbbreviations:
    """The Journal abbreviation singleton.

    This utility centralizes coBib's methodology of converting between full and abbreviated journal
    names. It implements the singleton pattern to enforce consistency across all modules.
    """

    _instance: Optional[JournalAbbreviations] = None
    """The singleton instance of this class."""

    _abbreviations: Dict[str, str] = {}
    """The parsed abbreviations."""

    _fullwords: Dict[str, str] = {}
    """The inverted, parsed abbreviations."""

    def __new__(cls) -> JournalAbbreviations:
        """Singleton constructor.

        This method gets called when accessing `JournalAbbreviations` and enforces the singleton
        pattern implemented by this class.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @staticmethod
    def load_abbreviations() -> None:
        """Load the list of abbreviations from the user configuration.

        By default, coBib does not ship with any abbreviations. Instead, the user is required to
        provide his own list via `config.utils.journal_abbreviations`. This setting must be a list
        of tuples formatted as: `("full name", "abbreviation")`.
        """
        for fullword, abbreviation in config.utils.journal_abbreviations:
            JournalAbbreviations._abbreviations[fullword] = abbreviation
            JournalAbbreviations._fullwords[abbreviation] = fullword
            JournalAbbreviations._fullwords[abbreviation.replace(".", "")] = fullword

    @staticmethod
    def check_existence(journal: str) -> bool:
        """Checks whether a journal name is present in the user's configuration.

        Args:
            journal: the journal name to check for.

        Returns:
            A boolean indicating whether this journal name is configured.
        """
        if not JournalAbbreviations._abbreviations:
            JournalAbbreviations.load_abbreviations()

        if journal in JournalAbbreviations._abbreviations:
            return True
        if journal in JournalAbbreviations._fullwords:
            return True

        msg = (
            f"'{journal}' was not found in your list of journal abbreviations! If you want to be "
            "able to automatically convert between full journal names and their abbreviated "
            "versions, consider adding a tuple like '(full journal name, abbreviation)' to the list"
            " stored under config.utils.journal_abbreviations"
        )
        LOGGER.warning(msg)
        return False

    @staticmethod
    def abbreviate(journal: str, dotless: bool = False) -> str:
        """Abbreviates the given journal name.

        Args:
            journal: the journal name to abbreviate.
            dotless: whether to return the abbreviated name without punctuation.

        Returns:
            The abbreviated journal name. If the provided journal name has no configured
            abbreviation, it will be returned unchanged. The `dotless` argument will have *no*
            effect!
        """
        if not JournalAbbreviations.check_existence(journal):
            return journal

        remove_punctuation: Callable[[str], str] = (
            lambda journal: journal.replace(".", "") if dotless else journal
        )

        if journal in JournalAbbreviations._fullwords:
            LOGGER.debug("'%s' is already abbreviated.", journal)
            return remove_punctuation(journal)

        new_journal = JournalAbbreviations._abbreviations.get(journal, journal)

        return remove_punctuation(new_journal)

    @staticmethod
    def elongate(journal: str) -> str:
        """Elongates the given journal name.

        Args:
            journal: the journal name to elongate.

        Returns:
            The elongated journal name. If the provided journal name has no configured elongation,
            it will be returned unchanged.
        """
        if not JournalAbbreviations.check_existence(journal):
            return journal

        if journal in JournalAbbreviations._abbreviations:
            LOGGER.info("'%s' is already elongated.", journal)
            return journal

        new_journal = JournalAbbreviations._fullwords.get(journal, journal)
        return new_journal
